/* This file is part of Mailfromd.
   Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <syslog.h>
#include "libmf.h"
#ifdef USE_SYSLOG_ASYNC
# include "syslog_async.h"
#endif

static int
vformat_logmsg(char **pbuf, size_t *psize, size_t off,
	       const char *fmt, va_list ap)
{
	char *buf = *pbuf;
	size_t buflen = *psize;
	int rc = 0;
	
	if (!buf) {
		if (buflen == 0)
			buflen = off + 512; /* Initial allocation */

		buf = calloc(1, buflen);
		if (buf == NULL)
			return 1;
	}

	for (;;) {
		ssize_t n = vsnprintf(buf + off, buflen - off, fmt, ap);
		if (n < 0 || n >= buflen
		    || !memchr(buf + off, '\0', n + 1)) {
			char *newbuf;
			size_t newlen = buflen * 2;
			if (newlen < buflen) {
				rc = 1;
				break;
			}
			newbuf = realloc(buf, newlen);
			if (newbuf == NULL) {
				rc = 1;
				break;
			}
			buflen = newlen;
			buf = newbuf;
		} else
			break;
	}

	if (rc) {
		if (!*pbuf) {
			/* We made first allocation, now free it */
			free (buf);
			buf = NULL;
			buflen = 0;
		}
	}

	*pbuf = buf;
	*psize = buflen;
	return rc;
}

static int
format_logmsg(char **pbuf, size_t *psize, size_t off,
	      const char *fmt, ...)
{
	int rc;
	
	va_list ap;
	va_start(ap, fmt);
	rc = vformat_logmsg(pbuf, psize, off, fmt, ap);
	va_end(ap);
	return rc;
}


void
syslog_default_open(const char *ident, int option, int facility)
{
	openlog(ident, option, facility);
}

void
syslog_default_close(void)
{
	closelog();
}

void
syslog_default_printer(int prio, const char *fmt, va_list ap)
{
#if HAVE_VSYSLOG
	vsyslog(prio, fmt, ap);
#else
	static char *buf = NULL;
	static size_t bufsize = 0;
	char sbuf[128];

	if (vformat_logmsg(&buf, &bufsize, 0, fmt, ap) == 0)
		syslog(prio, "%s", buf);
	else {
		vsnprintf(sbuf, sizeof sbuf, fmt, ap);
		syslog(prio, "%s", sbuf);
	}
#endif
}


#ifdef USE_SYSLOG_ASYNC
void
syslog_async_open(const char *ident, int option, int facility)
{
	openlog_async(ident, option, facility);
}

void
syslog_async_close(void)
{
	closelog_async();
}

void
syslog_async_printer(int prio, const char *fmt, va_list ap)
{
	vsyslog_async(prio, fmt, ap);
}

void
syslog_async_fdset(fd_set *fds)
{
	FD_SET(log_fd_async(), fds);
}
#endif

void
stderr_printer(int prio, const char *fmt, va_list ap)
{
	vfprintf(stderr, fmt, ap);
	fputc('\n', stderr);
}

void
stderr_fdset(fd_set *fds)
{
	FD_SET(fileno(stderr), fds);
}


static struct logger loggertab[] = {
	{ "stderr",
	  LOGF_STDERR,
	  NULL,
	  NULL,
	  stderr_printer,
	  stderr_fdset },
	{ "syslog",
	  0,
	  syslog_default_open,
	  syslog_default_close,
	  syslog_default_printer,
	  NULL },
#ifdef USE_SYSLOG_ASYNC
	{ "syslog:async",
	  0,
	  syslog_async_open,
	  syslog_async_close,
	  syslog_async_printer,
	  NULL },
#endif
	{ NULL }
};
static int current_logger;

int
logger_flags(int mask)
{
	return loggertab[current_logger].flags & mask;
}

int
logger_select(const char *name)
{
	int i;
	for (i = 0; loggertab[i].name; i++) {
		if (strcmp(loggertab[i].name, name) == 0) {
			current_logger = i;
			return 0;
		}
	}
	return -1;
}

void
logger_open(const char *ident, int option, int facility)
{
	struct logger *lp = loggertab + current_logger;
	if (lp->log_open)
		lp->log_open(ident, option, facility);
}

void
logger_close()
{
	struct logger *lp = loggertab + current_logger;
	if (lp->log_close)
		lp->log_close();
}

void
logger_fdset(fd_set *set)
{
	struct logger *lp = loggertab + current_logger;
	if (lp->log_fdset)
		lp->log_fdset(set);
}

void
vlogmsg(int prio, const char *fmt, va_list ap)
{
	loggertab[current_logger].log_print(prio, fmt, ap);
}

void
logmsg(int prio, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vlogmsg(prio, fmt, ap);
	va_end(ap);
}

static char *
format_log_si(const char *file, int line, const char *fun,
	      char *fmt, va_list ap)
{
	static char sbuf[512];
	static char *buf = NULL;
	static size_t bufsize = 0;
	
	if (format_logmsg(&buf, &bufsize, 0, "%s:%d:%s: ",
			  file, line, fun)
	    || vformat_logmsg(&buf, &bufsize, strlen(buf), fmt, ap)) {
		size_t off = snprintf(sbuf, sizeof(sbuf), "%s:%d:%s: ",
				      file, line, fun);
		vsnprintf(sbuf + off, sizeof(sbuf) - off, fmt, ap);
		return sbuf;
	}
	return buf;
}

static const char *debug_loc_file;
static const char *debug_loc_func;
static int debug_loc_line;

void
debug_locus(const char *file, int line, const char *fun)
{
	debug_loc_file = file;
	debug_loc_line = line;
	debug_loc_func = fun;
}

void
debug_vlog(char *fmt, va_list ap)
{
	if (debug_loc_file) {
		logmsg(LOG_DEBUG, "%s",
		       format_log_si(debug_loc_file, debug_loc_line,
				     debug_loc_func, fmt, ap));
		debug_loc_file = NULL;
	} else
		vlogmsg(LOG_DEBUG, fmt, ap);
}

void
debug_log(char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	debug_vlog(fmt, ap);
	va_end(ap);
}

